:- module(
  gpac,
  [
    %Commands
    compile_from_expression/3,
    compile_from_pivp/3,
    %Grammar
    polynomial/1,
    monomial/1,
    pode/1,
    pivp/1,
    %API
    test_2908/0,
    format_pivp/2
  ]).

:- doc('Biocham can also compile mathematical expressions and solutions of polynomial differential equations into biochemical reaction networks using only positive concentration values.').

:- devdoc('\\begin{todo} Implement the decomposition into at most binary reactions as an option.\\end{todo}').


compile_from_expression(Expr, Input, Output) :-
  biocham_command,
  type(Expr, arithmetic_expression),
  type(Input, name),
  type(Output, name),
  doc('
    creates a biochemical reaction network such that Output = Expr(Input). Use \'time\' as
    input to get Output = Expr(t).
  '),
  expression_to_PIVP(Expr, PIVP),
  pivp_compile(PIVP, Input, Output).

:- doc('
  \\begin{example} Compilation of the expression 4+time^2:\n
').
:- biocham(compile_from_expression(4 + id*id, time, output)).
:- biocham(list_model).
:- biocham(numerical_simulation(time:4)). %, method:msbdf)).
:- biocham(plot).
:- doc('
  \\end{example}
  ').

:- doc('
  \\begin{example} Compilation of the expression cos(time):\n
').
:- biocham(compile_from_expression(cos, time, f)).
:- biocham(list_model).
:- biocham(numerical_simulation(time:10)). %, method:msbdf)).
:- biocham(plot).
:- doc('
  \\end{example}
').


:- doc('
  \\begin{example} Compilation of the expression cos(x) and computation of cos(4):\n
').
:- biocham(compile_from_expression(cos, x, r)).
:- biocham(list_model).
:- biocham(present(x_p, 4)).
:- biocham(numerical_simulation(time:10)). %, method:msbdf)).
:- biocham(plot).
:- doc('
  \\end{example}
').

:- devdoc('\\begin{todo} share common subexpression\\end{todo}').

:- doc('
  \\begin{example} Compilation of a Hill function of order 2 as a function of time\n
').
:- biocham(compile_from_expression(id*id/(1+id*id), time, hill)).
:- biocham(list_model).
:- biocham(numerical_simulation(time:20)). % method:msbdf
:- biocham(plot(show:{hill_p,hill_m})).
:- biocham(plot(show:{hill_p,hill_m}, logscale:x)).
:- doc('
  \\end{example}
').



:- doc('One can also compile real valued functions defined as solutions of Polynomial Initial Value Problems (PIVP) using the following syntax:').


:- grammar(pivp).

pivp((X;Y)):-
  pivp(X),
  pivp(Y).

pivp((X,Y)):-
  number(X),
  pode(Y).

:- grammar(pode).

pode((d(X)/dt = Y)):-
  name(X),
  polynomial(Y).

:- grammar(polynomial).

polynomial(+ X):-
  polynomial(X).

polynomial(- X):-
  polynomial(X).

polynomial(X + Y):-
  polynomial(X),
  polynomial(Y).

polynomial(X - Y):-
  polynomial(X),
  monomial(Y).

polynomial(X):-
  monomial(X).

:- grammar(monomial).

monomial(X ^ Y):-
  name(X),
  number(Y).

monomial(X * Y):-
  monomial(X),
  monomial(Y).

monomial(X) :-
  number(X).

monomial(X) :-
  name(X).


compile_from_pivp(PIVP, Input, Output) :-
  biocham_command,
  type(PIVP, pivp),
  type(Input, name),
  type(Output, name),
  doc('
    compiles a PIVP into biochemical reactions. Use \'time\' as input
    to get output f(t), or use any other input x to get f(x) as output.
  '),
  format_pivp(PIVP, P),
  pivp_compile(P, Input, Output).

:- devdoc('\\begin{todo} Compilation of the cosine function as a function of time defined as the solution of a PIVP)\n
\\texttt{
 :- biocham(compile_from_pivp(((1,(d(y)/dt=z)); (0,(d(z)/dt= (-1*y)))), time, output)).
 :- biocham(list_model).
 :- biocham(numerical_simulation(time:10)).
 :- biocham(plot).
}
   \\end{todo}
   ').

test_2908 :-
%  biocham_command,
  exp_PIVP(EXP_PIVP),
  invert_PIVP(EXP_PIVP, INVEXP_PIVP),
  clear_model,
  compile_PIVP(INVEXP_PIVP, input, output).



:- devdoc('
  The internal data structure for PIVPs (Polynomial differential equations
  Initial Value Problems) is
  \\texttt{[ max(1,dimension), [ differential equations ] , [initial values] ]}

  The differential equations are represented by the list of lists (sums) of
  monomials (list of exponents) for each variable.

  \texttt{[ [ [coeff, [exponent x1,..., exponent xn]], ... other monomials for
  x1 differential... ], ... other differential functions for xi differential... ]}
').


pivp_compile(PIVP, Input, Output):-
    clear_model,
    get_option(fast_rate, Fast),
    set_parameter(fast, Fast),
  (
    Input = time
  ->
    compile_PIVP(PIVP, Input, Output)
  ;
    g_to_c_PIVP(PIVP, PIVPc, 0),
    assert(name_spec),
    compile_PIVP(PIVPc, Input, Output),
    retract(name_spec)
  ).



format_pivp(P, [N, Pode, Init]):-
  find_variables(P, L, Init),
  length(L, N), % should put 1 if no variable ?
  parse_pivp(P, L, Pode).

:- devdoc('
  \\command{format_pivp/2} encodes a PIVP syntax tree into the internal list
  representation for PIVPs.
').

find_variables((X;Y), L, Init):-
  find_variables(X, L1, I1),
  find_variables(Y, L2, I2),
  append(L1, L2, L),
  append(I1, I2, Init).

find_variables((I, (d(X)/dt=_)), [X], [I]).


parse_pivp((X;Y), L, Pode):-
  parse_pivp(X, L, P1),
  parse_pivp(Y, L, P2),
  append(P1, P2, Pode).

parse_pivp((_,(d(_)/dt=P)), L, Pol):-
  parse_polynomial(P, L, Pol).


parse_monomial(X ^ Y, L, M):-
  name(X),
  number(Y),
  make_monomial(L, X, Y, M).

parse_monomial(X * Y, L, M):-
  parse_monomial(X, L, M1),
  parse_monomial(Y, L, M2),
  append(M1, M2, M).

parse_monomial(X, _ , [X, []]) :-
  number(X).

parse_monomial(X, L, [1, M]) :-
  name(X),
  make_monomial(L, X, 1, M).



parse_polynomial(+ X, L, P):-
  parse_polynomial(X, L, P).

parse_polynomial(- X, L, [D, R]):-
  parse_polynomial(X, L, [C, R]),
  D is -C.

parse_polynomial(X + Y, L, R):-
  parse_polynomial(X, L, R1),
  parse_polynomial(Y, L, R2),
  append(R1, R2, R).

parse_polynomial(X - Y, L, [[D, M] | R1]):-
  parse_polynomial(X, L, R1),
  parse_monomial(Y, L, [C, M]),
  D is -C.

parse_polynomial(X, L, [M]):-
  parse_monomial(X, L, M).



make_monomial([], _, _, []).

make_monomial([X|L], X, Y, [Y|M]):-
  !,
  make_monomial(L, X, Y, M).

make_monomial([_|L], X, Y, [0|M]):-
  make_monomial(L, X, Y, M).





:- dynamic(input_name/1,
  output_name/1,
  name_spec/0).

compare_exponants([A1], [A2], C) :-
  (
    A1 < A2
  ->
    C = l
  ;
    A1 =:= A2
  ->
    C = e
  ;
    A1 > A2
  ->
    C = g
  ),
  !.

compare_exponants([A1|T1], [A2|T2], C) :-
  compare_exponants([A1], [A2], K),
  !,
  (
    K = e
  ->
    compare_exponants(T1, T2, C)
  ;
    C = K
  ).


add_exponants([], [], []) :- !.
add_exponants([A1|T1], [A2|T2], [A|T]) :-
  A is A1 + A2,
  add_exponants(T1, T2, T).


add_multivar([], P, P) :- !.
add_multivar(P, [], P) :- !.
add_multivar([[X1, E1]|T1], [[X2, E2]|T2], Sum) :-
  compare_exponants(E1, E2, C),
  ( C = l -> add_multivar(T1, [[X2, E2]|T2], T),
  Sum = [[X1, E1]|T] ;
  C = e -> add_multivar(T1, T2, T),
  X is X1 + X2,
  ( X =\= 0 -> Sum = [[X, E1]|T] ; Sum = T) ;
  C = g -> add_multivar([[X1, E1]|T1], T2, T),
  Sum = [[X2, E2]|T] ).

multiply_multivar(P, one, P) :- !.
multiply_multivar(one, P, P) :- !.
multiply_multivar(_, [], []) :- !.
multiply_multivar(P1, [M2|T2], P) :-
  multiply_multivar_monomial(P1, M2, P1M2),
  multiply_multivar(P1, T2, P1T2),
  add_multivar(P1M2, P1T2, P).

multiply_multivar_monomial([], _, []) :- !.
multiply_multivar_monomial([[X1, E1]|T1], [X2, E2], [[X, E]|T]) :-
  X is X1 * X2,
  add_exponants(E1, E2, E),
  multiply_multivar_monomial(T1, [X2, E2], T).

split([], [], []) :- !.
split([M], [M], []) :- !.
split([M1, M2|T], [M1|T1], [M2|T2]) :-
  split(T, T1, T2).

sort_multivar([], []) :- !.
sort_multivar([M], [M]) :- !.
sort_multivar(P, Ps) :-
  split(P, P1, P2),
  sort_multivar(P1, P1s),
  sort_multivar(P2, P2s),
  add_multivar(P1s, P2s, Ps).

translate_exponant(E, 0, E) :- !.
translate_exponant(E, N, [0|Ep]) :-
  Np is N - 1,
  translate_exponant(E, Np, Ep).

extend_exponant([], N, Z) :-
  translate_exponant([], N, Z), !.
extend_exponant([A|T], N, [A|ET]) :-
  extend_exponant(T, N, ET).

displace_exponant(E, Left, Right, Ed) :-
  translate_exponant(E, Left, Et),
  extend_exponant(Et, Right, Ed).

displace_multivar([], _, _, []) :- !.
displace_multivar([[X, E]|T], Left, Right, [[X, Ed]|Td]) :-
  displace_exponant(E, Left, Right, Ed),
  displace_multivar(T, Left, Right, Td).

displace_multivar_vector([], _, _, []) :- !.
displace_multivar_vector([P|T], Left, Right, [Pd|Td]) :-
  displace_multivar(P, Left, Right, Pd),
  displace_multivar_vector(T, Left, Right, Td).

multiply_multivar_vector([], _, []) :- !.
multiply_multivar_vector([P1|T1], P2, [P1P2|T]) :-
  multiply_multivar(P1, P2, P1P2),
  multiply_multivar_vector(T1, P2, T).

get_element([H|_], 1, H) :- !.
get_element([_|T], N, I) :-
  Np is N - 1,
  get_element(T, Np, I).

unitary(0, []) :- !.
unitary(N, [0|T]) :-
  Nxt is N - 1,
  unitary(Nxt, T).

power_multivar(0, [[_, E]|_], [[1, F]]) :-
  length(E, N), !,
  unitary(N, F).
power_multivar(N, P, Q) :-
  Np is N - 1,
  power_multivar(Np, P, P1),
  multiply_multivar(P, P1, Q).

scalar_multivar(0, _, []) :- !.
scalar_multivar(_, [], []) :- !.
scalar_multivar(X1, [[X2, E]|T2], [[X, E]|T]) :-
  X is X1 * X2,
  scalar_multivar(X1, T2, T).

compose_multivar_monomial([X, E], PL, P) :-
  compose_multivar_unitary(E, PL, P0),
  scalar_multivar(X, P0, P).

compose_multivar_unitary([], [], one) :- !.
compose_multivar_unitary([A|AT], [P|PT], Q) :-
  power_multivar(A, P, P0),
  compose_multivar_unitary(AT, PT, P1),
  multiply_multivar(P0, P1, Q).

compose_multivar([], _, []) :- !.
compose_multivar([[X, E]|T], PL, Q) :-
  compose_multivar_monomial([X, E], PL, P0),
  compose_multivar(T, PL, P1),
  add_multivar(P0, P1, Q).

compose_multivar_vector([], _, []) :- !.
compose_multivar_vector([P|PV], PL, [Q|QV]) :-
  compose_multivar(P, PL, Q),
  compose_multivar_vector(PV, PL, QV).

add_PIVP([D1, PV1, IV1], [D2, PV2, IV2], [D, PV, IV]) :-
  D1_p_1 is D1 + 1,
  displace_multivar_vector(PV1, 1, D2, PV1d),
  displace_multivar_vector(PV2, D1_p_1, 0, PV2d),
  get_element(PV1d, 1, P1d),
  get_element(PV2d, 1, P2d),
  get_element(IV1, 1, I1),
  get_element(IV2, 1, I2),

  add_multivar(P1d, P2d, P),
  I is I1 + I2,

  D is D1_p_1 + D2,
  append([[P], PV1d, PV2d], PV),
  append([[I], IV1, IV2], IV).

multiply_PIVP([D1, PV1, IV1], [D2, PV2, IV2], [D, PV, IV]) :-
  D1_p_1 is D1 + 1,
  Right1 is D1 + D2 - 1,
  Right2 is D2 - 1,
  displace_multivar_vector(PV1, 1, D2, PV1d),
  displace_multivar_vector(PV2, D1_p_1, 0, PV2d),
  get_element(PV1d, 1, P1d),
  get_element(PV2d, 1, P2d),
  get_element(IV1, 1, I1),
  get_element(IV2, 1, I2),
  displace_exponant([1], 1, Right1, E1),
  displace_exponant([1], D1_p_1, Right2, E2),
  multiply_multivar_monomial(P1d, [1, E2], Q1),
  multiply_multivar_monomial(P2d, [1, E1], Q2),
  add_multivar(Q1, Q2, Q),
  I is I1 * I2,
  D is D1_p_1 + D2,
  append([[Q], PV1d, PV2d], PV),
  append([[I], IV1, IV2], IV).

:- devcom('\\begin{todo} ^, log, compose PIVP\\end{todo}').

:- devcom('\\begin{todo}automatic simulation time horizon option for having stabilization on one component at epsilon from the middle (or two third) of the time horizon\\end{todo}').


invert_PIVP([D1, PV1, IV1], [D, PV, IV]) :-
  D is D1 + 1,

  displace_multivar_vector(PV1, 1, 0, PV1d),
  get_element(PV1d, 1, P1d),
  get_element(IV1, 1, I1),

  displace_exponant([2], 0, D1, E1),
  multiply_multivar_monomial(P1d, [-1, E1], Q),

  I is 1 / I1,
  append([[Q], PV1d], PV),
  append([[I], IV1], IV).


g_to_c_PIVP([D, PV, IV], [Dc, PVc, IVc], X) :-
  D_p_1 is D + 1,
  Dc is D + 2,
  displace_multivar_vector(PV, 2, 0, PVd),
  displace_exponant([1], 0, D_p_1, Ex),
  displace_exponant([1], 1, D, Egamma),

  X_m_Gamma = [[1, Ex], [-1, Egamma]],
  multiply_multivar_vector(PVd, X_m_Gamma, PVm),

  append([[[]], [X_m_Gamma], PVm], PVc),
  append([[X], [0], IV], IVc).


:- devcom('\\begin{todo}compilation option for not splitting the variables, assumed positive.\\end{todo}').

compile_PIVP([D, PV, IV], Input, Output) :-
  assert(input_name(Input)),
  assert(output_name(Output)),

  diff_list(D, Diff, D),
  compose_multivar_vector(PV, Diff, PVdiff),
  sign_split_multivar_vector(PVdiff, PVdiff_p, PVdiff_m),

  set_initial_vector(IV, 1),
  chemical_conversion(PVdiff_p, PVdiff_m, 1),
  anihilation_reactions(D),

  retract(input_name(Input)),
  retract(output_name(Output)).


diff_list(0, [], _) :- !.
diff_list(N, Diff, D) :-

  Left_p is 2*(N-1),
  Right_p is 2*(D-N)+1,
  Left_m is 2*N-1,
  Right_m is 2*(D-N),

  Np is N - 1,
  diff_list(Np, Diffp, D),

  displace_exponant([1], Left_p, Right_p, E_p),
  displace_exponant([1], Left_m, Right_m, E_m),
  P = [[1, E_p], [-1, E_m]],

  append([Diffp, [P]], Diff).

sign_split_multivar([], [], []) :- !.
sign_split_multivar([[X, E]|T], Plus, Minus) :-
  sign_split_multivar(T, T_p, T_m),
  ( X > 0 -> Plus = [[X, E]|T_p], Minus = T_m ;
    X < 0 -> Plus = T_p, Y is -X, Minus = [[Y, E]|T_m] ;
    Plus = T_p, Minus = T_m ).

sign_split_multivar_vector([], [], []) :- !.
sign_split_multivar_vector([P|PV], [Plus|PlusV], [Minus|MinusV]) :-
  sign_split_multivar(P, Plus, Minus),
  sign_split_multivar_vector(PV, PlusV, MinusV).


name_p_m(X,Xp,Xm):-
    atom_concat(X,'_p',Xp),
    atom_concat(X,'_m',Xm).

name_p_m(X,I,Xp,Xm):-
    atom_concat(X,I,XI),
    name_p_m(XI,Xp,Xm).


set_initial_vector([], _) :- !.
set_initial_vector([X|T], D) :-
    ( name_spec, D =:= 1 -> input_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_spec, D =:= 2 -> Z_p = g_p, Z_m = g_m ;
  name_spec, D =:= 3 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  (not name_spec), D =:= 1 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_p_m('z', D, Z_p,  Z_m) ),

  ( X > 0 ->  present([Z_p], X) ;
  X < 0 -> Y is -X, present([Z_m], Y) ; true ),
  Dp is D + 1,
  set_initial_vector(T, Dp).

anihilation_reactions(0) :- !.
anihilation_reactions(D) :-
    ( name_spec, D =:= 1 -> input_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_spec, D =:= 2 -> Z_p = g_p, Z_m = g_m ;
  name_spec, D =:= 3 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  (not name_spec), D =:= 1 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_p_m('z', D, Z_p,  Z_m) ),

  add_reaction(fast*Z_m*Z_p for Z_m+Z_p=>_),
  Dm is D - 1,
  anihilation_reactions(Dm).

chemical_conversion([], [], _) :- !.
chemical_conversion([Plus|PlusV], [Minus|MinusV], N) :-
    ( name_spec, N =:= 1 -> input_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_spec, N =:= 2 -> Z_p = g_p, Z_m = g_m ;
  name_spec, N =:= 3 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  (not name_spec), N =:= 1 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_p_m('z', N, Z_p,  Z_m) ),

  polynomial_to_reactions(Plus, Z_p),
  polynomial_to_reactions(Minus, Z_m),
  Nxt is N + 1,
  chemical_conversion(PlusV, MinusV, Nxt).

polynomial_to_reactions([], _) :- !.
polynomial_to_reactions([M|P], Y) :-
  monomial_to_reaction(M, Y),
  polynomial_to_reactions(P, Y).

monomial_to_reaction([X, E], Y) :-
  exponant_to_solution(E, 1, Sol, Kin),
  ( Sol = '_' -> add_reaction(X*Kin for _=>Y) ; add_reaction(X*Kin for _=[Sol]=>Y) ).

exponant_to_solution([], _, '_', 1) :- !.
exponant_to_solution([A_p, A_m|A], N, Sol, Kin) :-
  Np is N + 1,
  exponant_to_solution(A, Np, Sol0, Kin0),

  ( name_spec, N =:= 1 -> input_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_spec, N =:= 2 -> Z_p = g_p, Z_m = g_m ;
  name_spec, N =:= 3 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  (not name_spec), N =:= 1 -> output_name(Z), name_p_m(Z, Z_p,  Z_m) ;
  name_p_m('z', N, Z_p,  Z_m) ),

  ( A_p =:= 0 -> Sol1 = Sol0, Kin1 = Kin0 ;
    Sol0 = '_' -> Sol1 = A_p*Z_p, Kin1 = Z_p^A_p ; Sol1 = A_p*Z_p + Sol0, Kin1 = Z_p^A_p * Kin0 ),
    ( A_m =:= 0 -> Sol = Sol1, Kin = Kin1 ;
      Sol1 = '_' -> Sol = A_m*Z_m, Kin = Z_m^A_m ; Sol = A_m*Z_m + Sol1, Kin = Z_m^A_m * Kin1 ).


id_PIVP([1, [ [ [1, [0]] ] ], [0]]).

inv_PIVP([1, [ [ [-1, [2]] ] ], [1]]).

exp_PIVP([1, [ [ [1, [1]] ] ], [1]]).

invexp_PIVP([1, [ [ [-1, [1]] ] ], [1]]).

cos_PIVP([2, [ [[1, [0, 1]]] , [[-1, [1, 0]]] ], [1, 0]]).

sin_PIVP([2, [ [[1, [0, 1]]] , [[-1, [1, 0]]] ], [0, 1]]).


expression_to_PIVP(N, PIVP) :-
  number(N),
  !,
  PIVP = [1, [ [] ], [N]], !. 

expression_to_PIVP(id, PIVP) :-
  !,
  id_PIVP(PIVP).

expression_to_PIVP(exp, PIVP) :-
  !,
  exp_PIVP(PIVP).

expression_to_PIVP(cos, PIVP) :-
  !,
  cos_PIVP(PIVP).

expression_to_PIVP(sin, PIVP) :-
  !,
  sin_PIVP(PIVP).

expression_to_PIVP(E + F, PIVP) :-
  !,
  expression_to_PIVP(E, PIVP1),
  expression_to_PIVP(F, PIVP2),
  add_PIVP(PIVP1, PIVP2, PIVP).

expression_to_PIVP(E * F, PIVP) :-
  !,
  expression_to_PIVP(E, PIVP1),
  expression_to_PIVP(F, PIVP2),
  multiply_PIVP(PIVP1, PIVP2, PIVP).

expression_to_PIVP(1 / E, PIVP) :-
  !,
  expression_to_PIVP(E, PIVP1),
  invert_PIVP(PIVP1, PIVP), !.

expression_to_PIVP(E / F, PIVP) :-
  !,
  expression_to_PIVP(E, PIVP1),
  expression_to_PIVP(1 / F, PIVP2),
  multiply_PIVP(PIVP1, PIVP2, PIVP).

expression_to_PIVP(E, _) :-
  throw(error(illegal_expression(E), expression_to_PIVP)).


